from typing import Union

import math
import numpy as np

def mean(numbers: Union[np.ndarray, list]) -> float:
    """mean.
    Calculates the mean of a vector
    :param numbers:
    :type numbers: Union[np.ndarray, list]
    """
    return sum(numbers)/float(len(numbers))

def stdev(numbers: Union[np.ndarray, list]) -> float:
    """stdev.
    Calculates Standard Deviation of a vector
    :param numbers:
    :type numbers: Union[np.ndarray, list]
    :rtype: np.ndarray
    """
    avg = mean(numbers)
    variance = sum([pow(x-avg,2) for x in numbers])/float(len(numbers)-1)
    return math.sqrt(variance)

def gaussian_pdf(x: float, mean: float, stdev: float) -> float:
    """gaussian_pdf.
    Calculates the pdf value of x for a normal distribution (mean, stdev)
    :param x: one element of samples space
    :type x: float
    :param mean: mean of a normal distribution
    :type mean: float
    :param stdev: standard deviation of a normal distribution
    :type stdev: float
    :rtype: float
    """
    
    if stdev < 1e-3:
        if math.fabs(x-mean) < 1e-2:
            return 1.0
        else:
            return 0.0

    exponent = math.exp(-(math.pow(x-mean,2)/(2*math.pow(stdev, 2))))
    
    return (1 / (math.sqrt(2*math.pi) * stdev)) * exponent

def sigmoid(xvec: Union[np.ndarray, list]) -> np.ndarray:
    """sigmoid.
    Compute the sigmoid function
    :param xvec: a vector float values
    :type xvec: Union[np.ndarray, list]
    :rtype: np.ndarray
    """
    # Cap -xvec, to avoid overflow
    # Undeflow is okay, since it get set to zero
    xvec[xvec < -100] = -100

    vecsig = 1.0 / (1.0 + np.exp(np.negative(xvec)))

    return vecsig

def dsigmoid(xvec: np.ndarray) -> np.ndarray:
    """dsigmoid.
    Calculates gradient of standard sigmoid 1/(1+e^-x)
    :param xvec: a vector float values
    :type xvec: np.ndarray
    :rtype: np.ndarray
    """
    
    vecsig = sigmoid(xvec)
    
    return vecsig * (1 - vecsig)

def l2(vec: np.ndarray) -> float:
    """l2.
    Calculates l2 norm on vector vec
    :param vec: a vector float values
    :type vec: np.ndarray
    :rtype: float
    """
    return np.linalg.norm(vec)

def squared_l2(vec: np.ndarray) -> float:
    """squared_l2.
    Calculates squared l2 norm on vector vec
    :param vec: a vector float values
    :type vec: np.ndarray
    :rtype: float
    """
    return np.square(np.linalg.norm(vec))

def d_squared_l2(vec: np.ndarray) -> float:
    """squared_l2.
    Calculates gradient of squared l2 norm on a vector
    :param vec: a vector float values
    :type vec: np.ndarray
    :rtype: float
    """
    return vec

def l1(vec: np.ndarray) -> float:
    """l1.
    Calculates l1 norm on a vector
    :param vec: a vector float values
    :type vec: np.ndarray
    :rtype: float
    """
    return np.linalg.norm(vec, ord=1)

def d_l1(vec: np.ndarray) -> float:
    """d_l1.
    Calculates Subgradient of l1 norm on a vector
    :param vec: a vector float values
    :type vec: np.ndarray
    :rtype: float
    """
    
    grad = np.sign(vec)
    grad[abs(vec) < 1e-4] = 0.0
    return grad

def threshold_probs(probs: np.ndarray) -> np.ndarray:
    """threshold_probs.
    Converts probabilities to hard classification
    :param probs: a vector of probabilities
    :type probs: np.ndarray
    :rtype: np.ndarray
    """
    classes = np.ones(len(probs),)
    classes[probs < 0.5] = 0
    return classes

def logsumexp(a: np.ndarray) -> np.ndarray:
    """logsumexp.
    Compute the log of the sum of exponentials of input elements.
    Modified scipys logsumpexp implemenation for this specific situation
    :param a: a vector with dx1 dimensions
    :type a: np.ndarray
    :rtype: np.ndarray
    """

    awithzero = np.hstack((a, np.zeros((len(a),1))))
    maxvals = np.amax(awithzero, axis=1)
    aminusmax = np.exp((awithzero.transpose() - maxvals).transpose())

    # suppress warnings about log of zero
    with np.errstate(divide='ignore'):
        out = np.log(np.sum(aminusmax, axis=1))

    out = np.add(out,maxvals)

    return out

def update_dictionary_items(dict1: dict, dict2: dict) -> dict:
    """update_dictionary_items.
    Replace any common dictionary items in dict1 with the values in dict2
    There are more complicated and efficient ways to perform this task,
    but we will always have small dictionaries, so for our use case, this simple
    implementation is acceptable.
 
    :param dict1: first input dictionary
    :type dict1: dict
    :param dict2: second input dictionary
    :type dict2: dict
    :rtype: dict
   """

    if dict2 is None:
        return dict1
    for k in dict1:
        if k in dict2:
            dict1[k]=dict2[k]

    return dict1


def check_for_nans_in_data(X: np.ndarray, y: np.ndarray) -> None:
    """check_for_nans_in_data.
    Will print out nan vales in the input numpy arrays
    :param X:
    :type X: np.ndarray
    :param y:
    :type y: np.ndarray
    :rtype: None
    """
    if np.any(np.isnan(X)):
        print('X has nans!')
        for i in range(X.shape[0]):
            ind = np.isnan(X[i, :])
            if np.any(ind):
                print('X has nans at sample ' + str(i) + ' with y = ' + str(y[i]))

